//
// Find.cpp : Find/Replace
//

#include "stdafx.h"
#include "TinyPad.h"
#include <commctrl.h>
#include <aygshell.h>
#include <sipapi.h>

#include <commdlg.h>

#include "find.h"

static PFINDSTR		m_pfs;
static int			m_dlgmode;

/*
 *	FindProc
 */
static LRESULT CALLBACK FindProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	SHINITDLGINFO shidi;
	int state;
	TCHAR tbuf[MAX_PATH];

	switch (message)
	{
		case WM_INITDIALOG:
			SetDlgItemText(hDlg, IDC_FINDTEXT, m_pfs->aFindStr);
			if (m_dlgmode == IDD_REPLACE)
				SetDlgItemText(hDlg, IDC_REPLTEXT, m_pfs->aReplStr);

			state = m_pfs->CaseSensitive ? BST_CHECKED : BST_UNCHECKED;
			SendDlgItemMessage(hDlg, IDC_CASE_SENSITIVE, BM_SETCHECK, state, 0);

			// Create a Done button and size it.  
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
			shidi.hDlg = hDlg;
			SHInitDialog(&shidi);
			return TRUE; 

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
			case IDC_REPLALL:
				m_pfs->nFindStr = GetDlgItemText(hDlg, IDC_FINDTEXT, m_pfs->aFindStr, m_pfs->nFindStr);
				if (m_dlgmode == IDD_REPLACE)
					m_pfs->nReplStr = GetDlgItemText(hDlg, IDC_REPLTEXT, m_pfs->aReplStr, m_pfs->nReplStr);

				m_pfs->CaseSensitive = IsDlgButtonChecked(hDlg, IDC_CASE_SENSITIVE) == BST_CHECKED;

				if (m_pfs->nFindStr == 0 || m_pfs->nReplStr == 0)
					break;

				if (!ReplAll(m_pfs)) {
					wsprintf(tbuf, TEXT("\"%s\"%s"), m_pfs->aFindStr, szNotFound);
					MessageBox(hDlg, tbuf, szTitle, MB_ICONINFORMATION | MB_OK);
				}

				EndDialog(hDlg, (WPARAM)FALSE);
				return TRUE;
			case IDOK:
				m_pfs->nFindStr = GetDlgItemText(hDlg, IDC_FINDTEXT, m_pfs->aFindStr, m_pfs->nFindStr);
				if (m_dlgmode == IDD_REPLACE)
					m_pfs->nReplStr = GetDlgItemText(hDlg, IDC_REPLTEXT, m_pfs->aReplStr, m_pfs->nReplStr);

				m_pfs->CaseSensitive = IsDlgButtonChecked(hDlg, IDC_CASE_SENSITIVE) == BST_CHECKED;

				EndDialog(hDlg, (WPARAM)TRUE);
				return TRUE;
			case IDCANCEL:
				EndDialog(hDlg, (WPARAM)FALSE);
				return TRUE;
			}
			break;
	}
    return FALSE;
}

/*
 *	GetFindStr
 */
BOOL GetFindStr(PFINDSTR pfs)
{
	m_pfs = pfs;
	m_dlgmode = IDD_FIND;

	return DialogBox(pfs->hInst, (LPCTSTR)m_dlgmode, pfs->hWnd, (DLGPROC)FindProc);
}

/*
 *	GetReplStr
 */
BOOL GetReplStr(PFINDSTR pfs)
{
	m_pfs = pfs;
	m_dlgmode = IDD_REPLACE;

	return DialogBox(pfs->hInst, (LPCTSTR)m_dlgmode, pfs->hWnd, (DLGPROC)FindProc);
}

/*
 *	_StrPrev
 */
PTCHAR _StrPrev(PTCHAR pstart, PTCHAR plast, PFINDSTR pfs)
{
	PTCHAR psrch = pfs->aFindStr;
	int nsrch = pfs->nFindStr;

	if (pfs->CaseSensitive) {
		while (pstart >= plast) {
			if (wcsncmp(pstart, psrch, nsrch) == 0)
				return pstart;
			pstart--;
		}
	} else {
		while (pstart >= plast) {
			if (_wcsnicmp(pstart, psrch, nsrch) == 0)
				return pstart;
			pstart--;
		}
	}

	return NULL;
}

/*
 *	_StrNext
 */
PTCHAR _StrNext(PTCHAR pstart, PTCHAR plast, PFINDSTR pfs)
{
	PTCHAR psrch = pfs->aFindStr;
	int nsrch = pfs->nFindStr;

	if (pfs->CaseSensitive) {
		while (pstart < plast) {
			if (wcsncmp(pstart, psrch, nsrch) == 0)
				return pstart;
			pstart++;
		}
	} else {
		while (pstart < plast) {
			if (_wcsnicmp(pstart, psrch, nsrch) == 0)
				return pstart;
			pstart++;
		}
	}

	return NULL;
}

/*
 *	FindStr
 */
BOOL FindStr(PFINDSTR pfs)
{
	int iStart, iEnd;
	int count;
	int bufsize;
	PTCHAR pbuf = NULL;
	PTCHAR pfind;
	BOOL res = FALSE;

	if (pfs->nFindStr == 0)
		goto out;

	// Get current selection
	SendMessage(pfs->hWnd, EM_GETSEL, (WPARAM)&iStart, (LPARAM)&iEnd);

	// Load text
	count = SendMessage(pfs->hWnd, WM_GETTEXTLENGTH, 0, 0);
	bufsize = (count + 1) * sizeof(TCHAR);	// Including NUL
	if (!(pbuf = (PTCHAR)LocalAlloc(LMEM_FIXED, bufsize)))
		goto out;

	SendMessage(pfs->hWnd, WM_GETTEXT, count + 1, (LPARAM)pbuf);
	pbuf[count] = 0;	// NUL terminate

	// Search text
	if (pfs->BackSearch) {
		if (iStart < pfs->nFindStr)
			goto out;
		pfind = _StrPrev(&pbuf[iStart - pfs->nFindStr], pbuf, pfs);
	} else
		pfind = _StrNext(&pbuf[iEnd], &pbuf[count], pfs);

	if (!pfind)
		goto out;

	// Find the text
	iStart = (int)(pfind - pbuf);
	iEnd = iStart + pfs->nFindStr;
	SendMessage(pfs->hWnd, EM_SETSEL, iStart, iEnd);
	SendMessage(pfs->hWnd, EM_SCROLLCARET, 0, 0);

	res = TRUE;

out:
	if (pbuf)
		LocalFree(pbuf);

	return res;
}

/*
 *	ReplStr
 */
BOOL ReplStr(PFINDSTR pfs)
{
	int iStart, iEnd;
	int count, selcount;
	int bufsize;
	PTCHAR pbuf = NULL;
	BOOL res = FALSE;

	if (pfs->nFindStr == 0)
		goto out;

	// Get current selection
	SendMessage(pfs->hWnd, EM_GETSEL, (WPARAM)&iStart, (LPARAM)&iEnd);
	selcount = iEnd - iStart;
	if (selcount != pfs->nFindStr)
		goto out;

	// Load text
	count = SendMessage(pfs->hWnd, WM_GETTEXTLENGTH, 0, 0);
	bufsize = (count + 1) * sizeof(TCHAR);	// Including NUL
	if (!(pbuf = (PTCHAR)LocalAlloc(LMEM_FIXED, bufsize)))
		goto out;

	SendMessage(pfs->hWnd, WM_GETTEXT, count + 1, (LPARAM)pbuf);
	pbuf[count] = 0;	// NUL terminate

	// Check current selection is matched as find string
	if (pfs->CaseSensitive) {
		if (wcsncmp(&pbuf[iStart], pfs->aFindStr, selcount) != 0)
			goto out;
	} else {
		if (_wcsnicmp(&pbuf[iStart], pfs->aFindStr, selcount) != 0)
			goto out;
	}

	// Replace the text
	SendMessage(pfs->hWnd, EM_REPLACESEL, TRUE, (LPARAM)pfs->aReplStr);
	SendMessage(pfs->hWnd, EM_SCROLLCARET, 0, 0);

	res = TRUE;

out:
	if (pbuf)
		LocalFree(pbuf);

	return res;
}

/*
 *	ReplAll
 */
BOOL ReplAll(PFINDSTR pfs)
{
	int ibuf;
	int idst;
	int ifound;
	int ilastchange;
	int count;
	int dstcount;
	int cnt;
	int bufsize;
	int dstsize;
	PTCHAR pbuf = NULL;
	PTCHAR pdst = NULL;
	PTCHAR pfind;
	BOOL res = FALSE;
	BOOL found;

	if (pfs->nFindStr == 0)
		goto out;

	// Load text
	count = SendMessage(pfs->hWnd, WM_GETTEXTLENGTH, 0, 0);
	bufsize = (count + 1) * sizeof(TCHAR);	// Including NUL
	if (!(pbuf = (PTCHAR)LocalAlloc(LMEM_FIXED, bufsize)))
		goto out;

	SendMessage(pfs->hWnd, WM_GETTEXT, count + 1, (LPARAM)pbuf);
	pbuf[count] = 0;	// NUL terminate

	// Allocate dest buffer
	if (pfs->nFindStr > pfs->nReplStr)
		dstsize = bufsize + sizeof(TCHAR);
	else
		dstsize = MulDiv(bufsize, pfs->nReplStr, pfs->nFindStr) + sizeof(TCHAR);
	if (!(pdst = (PTCHAR)LocalAlloc(LMEM_FIXED, dstsize)))
		goto out;
	dstcount = dstsize / sizeof(TCHAR);

	// Find string and replace all
	idst = 0;
	found = FALSE;
	for (ibuf = 0; ibuf < count; ) {
		pfind = _StrNext(&pbuf[ibuf], &pbuf[count], pfs);
		if (!pfind)
			break;
		found = TRUE;

		// Fill unchanged characters to dest buffer
		ifound = (int)(pfind - pbuf);
		cnt = ifound - ibuf;
		if (cnt > 0) {
			if (idst + cnt > dstcount)
				goto out;	// buffer overflow
			memcpy(&pdst[idst], &pbuf[ibuf], cnt * sizeof(TCHAR));
			ibuf += cnt;
			idst += cnt;
		}

		// Replace it
		if (idst + pfs->nReplStr > dstcount)
			goto out;	// buffer overflow
		ilastchange = idst;
		memcpy(&pdst[idst], pfs->aReplStr, pfs->nReplStr * sizeof(TCHAR));
		ibuf += pfs->nFindStr;
		idst += pfs->nReplStr;
	}
	if (!found)
		goto out;	// Not found. No change

	// Fill remain text
	if (ibuf < count) {
		cnt = count - ibuf;
		if (idst + cnt + 1 > dstcount)
			goto out;	// buffer overflow
		memcpy(&pdst[idst], &pbuf[ibuf], cnt * sizeof(TCHAR));
		ibuf += cnt;
		idst += cnt;
		pdst[idst++] = 0;	// NUL
	}

	// Replace the text
	SendMessage(pfs->hWnd, EM_SETSEL, 0, -1);
	SendMessage(pfs->hWnd, EM_REPLACESEL, 0, (LPARAM)pdst);

	// Move cursor to last changed position
	SendMessage(pfs->hWnd, EM_SETSEL, ilastchange, ilastchange + pfs->nReplStr);
	SendMessage(pfs->hWnd, EM_SCROLLCARET, 0, 0);

	res = TRUE;

out:
	if (pdst)
		LocalFree(pdst);
	if (pbuf)
		LocalFree(pbuf);

	return res;
}

// vi: ts=4 sw=4
